/*-
 * $Id: xgridagent-profile.c,v 1.2 2004/06/22 03:45:42 dccote Exp $
 *
 * See the file LICENSE for redistribution information. 
 * If you have not received a copy of the license, please contact CodeFactory
 * by email at info@codefactory.se, or on the web at http://www.codefactory.se/
 * You may also write to: CodeFactory AB, SE-903 47, Ume, Sweden.
 *
 * Copyright (c) 2002 Jonas Borgstrm <jonas@codefactory.se>
 * Copyright (c) 2002 Daniel Lundin   <daniel@codefactory.se>
 * Copyright (c) 2002 CodeFactory AB.  All rights reserved.
 */

#include <librr/rr.h>
#include <string.h>

#include "xgridagent-profile.h"
#include "xgridagent.h"

#define PRO_XGRIDAGENT_URI		"http://www.apple.com/beep/xgrid/controller/agent"

#define EMPTY_RPY "Content-Type: application/beep+xml\r\n\r\n<null />"

static gboolean frame_available (RRChannel *channel, RRFrame *frame, 
				 GError **error);
extern RRSimple * simple;

static void
rr_simple_init (GObject *object)
{
}

static void
rr_simple_class_init (GObjectClass *klass)
{
	RRChannelClass *channel_class = (RRChannelClass *)klass;

	channel_class->frame_available = frame_available;
}

GType 
rr_simple_get_type (void)
{
	static GType rr_type = 0;

	if (!rr_type) {
		static GTypeInfo type_info = {
			sizeof (RRSimpleClass),
			NULL,
			NULL,
			(GClassInitFunc) rr_simple_class_init,
			NULL,
			NULL,
			sizeof (RRSimple),
			16,
			(GInstanceInitFunc) rr_simple_init
		};
		rr_type = g_type_register_static (RR_TYPE_CHANNEL, "RRSimple", 
						  &type_info, 0);

		rr_channel_set_uri (rr_type, PRO_XGRIDAGENT_URI);
	}
	return rr_type;
}

static gboolean
frame_available (RRChannel *channel, RRFrame *frame, GError **error)
{
    char messageString[100];
	if (frame->type == RR_FRAME_TYPE_MSG) {
		MyFrame * mf;
        char *payload;
        char * message;
        long size;
        
		mf = NewNode(simple->mHeadFrame, &simple->mLinkedListLock);
		
		if ( (mf->buf = malloc(frame -> size+1)) != NULL) {
			mf->size = frame -> size;
			mf->msg = frame -> msgno;
			memcpy (mf->buf, frame-> payload, frame -> size);
			mf->buf[frame->size]=0;
			sprintf(messageString,"Copying received frame to buffer (%d bytes)",mf->size);
			LogMessage(messageString, kDebugBEEP);
		}
			
		if (frame -> more ) {
		  return TRUE;
		}
		

		/* Send a reply */
		RRMessage *rpy;

		rpy = rr_message_static_new (RR_FRAME_TYPE_RPY, EMPTY_RPY, 
					     strlen(EMPTY_RPY), FALSE);
		/* the reply has to have the same msgno as the MSG */
		rpy->msgno = frame->msgno;
		rr_channel_send_message (channel, rpy, error);
    
	    message = AggregateAndRemoveMSG(simple->mHeadFrame, frame-> msgno, &simple->mLinkedListLock);
	    xgridagent_ProcessBeepMSG(message,  strlen(message));

        return TRUE;
	}
	else if (frame->type == RR_FRAME_TYPE_RPY) {
		return TRUE;
	}
	else {
		/* Simple error handling for unexpected frames. */
		g_set_error (error, RR_BEEP_ERROR, RR_BEEP_CODE_SYNTAX_ERROR,
			     "Unexpected frame received.");
		return FALSE;
	}
}

RRSimple *
rr_simple_start (RRConnection *connection, GError **error)
{
	RRSimple *simple;

	g_return_val_if_fail (RR_IS_CONNECTION (connection), NULL);

	if ((simple = (RRSimple *)rr_connection_start (connection, NULL, 
						    RR_TYPE_SIMPLE, NULL, 
						    error)) == NULL)
		return NULL;

	return simple;
}

/*

gboolean
rr_simple_do_stuff (RRSimple *simple, GError **error)
{
	RRMessage *msg;

	g_return_val_if_fail (RR_IS_SIMPLE (simple), FALSE);

	msg = rr_message_static_new (RR_FRAME_TYPE_MSG, MSG_STRING, 
				     strlen (MSG_STRING), FALSE);

	return rr_channel_send_message (RR_CHANNEL (simple), msg, error);
}
*/

static RRFrame *
get_frame (RRMessage *message, gsize max_size)
{
  RRMessageStatic *stat = RR_MESSAGE_STATIC(message);
  RRFrame *frame;
  gboolean more;
  gint len;

  g_assert (RR_IS_MESSAGE (stat));

  len = MIN (max_size, stat->payload_len - stat->offset);
  more = (max_size < stat->payload_len - stat->offset);

  frame = rr_frame_new (message->type, message->channel->id, more, 
			message->msgno,
			len, message->ansno, 
			(guint8 *)stat->payload + stat->offset, 
			FALSE);

  rr_frame_reference_message (frame, message);

  stat->offset += len;

  return frame;
}


gboolean
xgridagent_SengMSG(RRSimple *simple, GError **error, char *ibuf, int ilen)         
{
	RRMessage *msg;
	long windowSize;
	gboolean err = FALSE;
	gsize max_size;
	gboolean more;
	gint len;

	LogMessage("Sending message",kDebug);
	LogMessage(ibuf,kDebugBEEPWithMessages);

	g_return_val_if_fail (RR_IS_SIMPLE (simple), FALSE);

	windowSize = rr_channel_get_window_size(RR_CHANNEL (simple));
	msg = rr_message_static_new (RR_FRAME_TYPE_MSG, ibuf, 
				     ilen, FALSE);
	
	/*
	  This apparently does not solve my "hang" problem:
	  I thought that, following, http://xml.resource.org/public/rfc/html/rfc3081.html
	  I could break up the message in smaller frames and that the controller would
	  digest it fine, but it does not seem to work. I also thought that send_message()
	  would actually take care of that but it does not appear to.

	  Bottom line is: if the results are larger than the advertised window of the 
	  controller (appears to be around 15k in my situation), then the code will hang
	  because the controller's window is full, and the present code does not do anything 
	  about it.

	  This must be fixed, but I don't know how.

	RRMessageStatic *stat = RR_MESSAGE_STATIC(msg);
	g_assert (RR_IS_MESSAGE (stat));

	stat->offset = 0;

	max_size = windowSize/2;

	if ( ilen >= max_size) {
	  RRFrame* frame;
	  do {
	    len = MIN (max_size, stat->payload_len - stat->offset);
	    more = (max_size < stat->payload_len - stat->offset);
	    
	    frame = rr_frame_new (msg->type, RR_CHANNEL (simple)->id, more,
				  RR_CHANNEL (simple)->msgno,
				  len, msg->ansno,
				  (guint8 *)stat->payload + stat->offset, 
				  FALSE);
	    
	    err = rr_channel_send_frame (RR_CHANNEL (simple), RR_FRAME(frame), error);
	    //	    rr_frame_reference_message (RR_FRAME(frame), msg);
	    stat->offset += len;

	  } while ( more );
	} else {

	err = rr_channel_send_message (RR_CHANNEL (simple), msg, error);

	}
	*/

	err = rr_channel_send_message (RR_CHANNEL (simple), msg, error);

	return err;

}

gboolean
rr_simple_close (RRSimple *simple, GError **error)
{
	g_return_val_if_fail (RR_IS_SIMPLE (simple), FALSE);

	if (!rr_channel_close (RR_CHANNEL (simple), RR_BEEP_CODE_SUCCESS, 
			       NULL, "bye", error))
		return FALSE;
	
	g_object_unref (G_OBJECT (simple));

	return TRUE;
}
