#include <librr/rr.h>
#include "xgridagent-profile.h"
#include "xgridagent.h"
#include <sys/stat.h>
#include <pthread.h>
#include <semaphore.h>
#include <config.h>
#include <fcntl.h>
#include <errno.h>

static char gMessageString[MAX_MEDIUM_BUFFER];

Agent gAgent;
RRSimple *simple = NULL;


RRSimple*
GetSimple()
{
  return simple;
}

int
main (gint argc, gchar **argv)
{
	GError *error = NULL;
	RRConnection *connection;
	RRProfileRegistry *profreg;

    if (argc != 2) {
        printf("Usage: xgridagent ip_address\n");
        return 1;
    }

	/* Initialize roadrunner */
	if (!rr_init (NULL, NULL, &error))
		g_error ("rr_init failed: %s\n", error->message);

	/* Tell roadrunner which profiles we want to support */
	profreg = rr_profile_registry_new ();
	rr_profile_registry_add_profile (profreg, RR_TYPE_SIMPLE, NULL);

	/* Create a connection object */
	if ((connection = rr_tcp_connection_new (profreg, argv[1], 
						 4111, &error)) == NULL)
		g_error ("connection failed: %s\n", error->message);

	if ((simple = rr_simple_start (connection, &error)) == NULL)
		g_error ("rr_simple_start failed: %s\n", error->message);

	if (!xgridagent (simple, &error))
		g_error ("rr_simple_do_stuff failed: %s\n", error->message);

	if (!rr_simple_close (simple, &error))
		g_error ("rr_simple_close failed: %s\n", error->message);

	/* Close the connection */
	if (!rr_connection_disconnect (connection, &error))
		g_error ("disconnect failed: %s\n", error->message);

	/* Stop the main-loop */
	if (!rr_exit (&error))
		g_error ("rr_exit failed: %s\n", error->message);

	return 0;
}


int
XgridInitAgent()
{

	xmlDocPtr doc;
	char buf[MAX_SMALL_BUFFER];
	struct hostent * hostStruct;
	int err;
	
	/* Initialize Agent parameters */
	gAgent.mCurrentTaskCount = 0;
	gAgent.mAgentStatus = kAgentAvailable;

	FILE* cookie;
	if ( (cookie = fopen("cookie","r")) == NULL) {
	  strcpy(gAgent.mAgentCookie,"");
	} else {
	  fscanf(cookie,"%s", gAgent.mAgentCookie);
	  fclose(cookie);
	}
	gAgent.mCurrentCPUPower = 0;
	gAgent.mSerialTaskRef = 2;
	gAgent.mMessageLevel = 2;
	/* Initialize thread stuff */
	gAgent.mPendingTasks = sem_open("pending",O_CREAT, 0777,0);
	if (gAgent.mPendingTasks == SEM_FAILED) {
	  sprintf(buf,"sem_open() failed with error number %d.  Trying sem_init()",errno);
	  LogMessage(buf,kDebug);
	  gAgent.mPendingTasks = malloc(sizeof(sem_t));
	  if ( SEM_INIT(gAgent.mPendingTasks,0) ) {
	    sprintf(buf,"Semaphore failed with error number %d",errno);
	    LogMessage(buf,kDebug);
	  } else {
	    LogMessage("Previous error was harmless", kDebug);
	  }
	}
    
	LogMessage("Initialized pending tasks", kDebug);
	pthread_mutex_init(&gAgent.mXgridQueueLocked, NULL);
	
	/* Load configuration document*/
    doc = xmlParseFile("xgrid.config.xml");
    if (doc == NULL) {
		return kCantReadConfigFile;
    }
	
	if ( ! extractSingleXPath(doc, "/plist/dict/string[preceding-sibling::key[position()=1 and text()='MaxTaskCount']]/text()", buf) ) {
		gAgent.mMaxTaskCount = atol(buf);
		if (gAgent.mMaxTaskCount <= 0) {
			return kMustHaveTaskCountPositive;
		}
	} else {
		return kCouldNotReadMaxTaskCount;
	}	
	
	if ( (gAgent.mTaskQueue = malloc(sizeof(Task) * gAgent.mMaxTaskCount)) == NULL) {
		return kOutOfMemoryErr;
	}
	
	if ( ! extractSingleXPath(doc, "/plist/dict/string[preceding-sibling::key[position()=1 and text()='ServiceName']]/text()", buf) ) {
		strcpy(gAgent.mAgentName, buf);
	} else {
		return kCouldNotReadServiceName;
	}
	
	gethostname(gAgent.mHostname,MAX_SMALL_BUFFER);

/*
  I could not figure out a portable way to do this.

	if ( (hostStruct = gethostbyname (gAgent.mHostname)) != NULL ) {
		strcpy(gAgent.mAddress, hostStruct->h_addr);
	} else {
		return kCouldNotGetAddress;
	}
  Since it's not essential, this is what I do:
*/
	strcpy(gAgent.mAddress, "Unknown");

	if ( ! extractSingleXPath(doc, "/plist/dict/string[preceding-sibling::key[position()=1 and text()='MaximumCPUPower']]/text()", buf) ) {
		gAgent.mMaximumCPUPower = atol(buf);
		if (gAgent.mMaxTaskCount <= 0) {
			return kMustHaveCPUPowerPositive;
		}
	} else {
		return kCouldNotReadCPUPower;
	}	

	if ( ! extractSingleXPath(doc, "/plist/dict/string[preceding-sibling::key[position()=1 and text()='MessageLevel']]/text()", buf) ) {
		gAgent.mMessageLevel = atol(buf);
		if (gAgent.mMessageLevel < 0) {
			return kMustHavePositiveMessageLevel;
		}
	} else {
		gAgent.mMessageLevel = kNormal;
	}	
	
	
	pthread_mutex_init(&simple->mLinkedListLock, NULL);
	simple->mHeadFrame = NewNode(NULL,&simple->mLinkedListLock);



	return kNoErr;
}

int ReportError(int inErr)
{
	char buf[MAX_MEDIUM_BUFFER];

	if (inErr <= kMustHavePositiveMessageLevel)
		sprintf(buf, "Error[%d]: %s", inErr, gErrorStrings[inErr]);
	else
		sprintf(buf, "Error[%d]: (no text defined for this message)", inErr);
		
	LogMessage(buf, kNormal);

	return inErr;
}

void LogMessage(char* inMsg, long inLevel)
{
	if (inLevel <= gAgent.mMessageLevel) {
		static char s1[MAX_SMALL_BUFFER];
        static char s2[MAX_SMALL_BUFFER];
        struct tm *date;
        time_t now = time(NULL);
        date = localtime(&now);
		
        strftime(s1, MAX_SMALL_BUFFER, "%x %H:%M:%S",date);
#if HAVE_UNISTD_H
        sprintf(s2," [%d]", getpid());
		strcat(s1,s2);
#endif
		
		fprintf(stderr, "%s %s\n", s1, inMsg);
		fflush(stderr);
	}
}


int
xgridagent (RRSimple* simple, GError** error) {
	
    int err = kNoErr;
	int status = 0;
    pthread_t thread;
	
	xmlInitParser();
	if ( (err = XgridInitAgent()) != kNoErr) {
		return ReportError(err);
	}
	
    if ( (err = pthread_create( &thread, NULL, (void*)&SimpleWatchThread, NULL)) != kNoErr) {
		return ReportError(kCouldNotStartMainThread);
	}
	
	XgMessage msg;
	RegistrationRequestMessage(&msg);
	
    if ((err = xgridagent_SengMSG (simple, error, msg.mStringRepresentation, strlen(msg.mStringRepresentation))) < 0) {
        return ReportError(kCouldNotSendRegistration);
    }
	
	pthread_join( thread, NULL);
	
    return err;
}


void* SimpleWatchThread(void * inPtr)
{
    long i;
    Task* task = NULL;
    pthread_t threads[MAX_THREADS];
    int res[MAX_THREADS];
    long count = 0;
    
    while (1) {
        SEM_WAIT(gAgent.mPendingTasks);
        LogMessage("Task queue modified.", kVerbose);
    	
		MUTEX_LOCK(&gAgent.mXgridQueueLocked);
		
		for (i = 0; i < gAgent.mCurrentTaskCount; i++) {
			task = gAgent.mTaskQueue[i];

			if (task->mStatus == kTaskMustStart) {
				MUTEX_LOCK(&task->mLock);
				task->mStatus = kTaskRunning;
				sprintf(gMessageString,"Starting new Xgrid task thread for task #%d",task->mTaskRef);
				LogMessage(gMessageString, kVerbose);
				res[count] = pthread_create( &threads[count], NULL, (void*)&XgridTaskMonitorThread, task);
				count++;
				MUTEX_UNLOCK(&task->mLock);
				break;
			} else if (task->mStatus == kTaskMustCancel) {
				sprintf(gMessageString,"Cancelling Xgrid task thread for task #%d",task->mTaskRef);
				LogMessage(gMessageString, kVerbose);
				MUTEX_LOCK(&task->mLock);
				RemoveXgridTaskFromQueue(task->mTaskRef);
				MUTEX_UNLOCK(&task->mLock);
				DeleteXgridTask(task);
				break;
			}
			

		}
		
		MUTEX_UNLOCK(&gAgent.mXgridQueueLocked);
		
    }
}

void*
XgridTaskMonitorThread(void* inPtr)
{
    
    pthread_t thread;
    int err;
    GError *error = NULL;
    
    Task* task = (Task*)inPtr;
    
    if ( (err = pthread_create( &thread, NULL, (void*)&XgridTaskThread, task) ) != kNoErr ) {
		ReportError(kCouldNotStartTaskThread);
		return NULL;
	}
		  
    while (1) {
      MUTEX_LOCK(&task->mLock);
      if (SEM_TRYWAIT(task->mTaskDone)) {
            // Not done yet
            XgMessage msg;
            TaskStatusNotificationMessage(&msg, task);
            MUTEX_UNLOCK(&task->mLock);
		
            if ( (err = xgridagent_SengMSG (simple, &error, msg.mStringRepresentation, strlen(msg.mStringRepresentation)) ) < 0) {
                ReportError(kCouldNotSendStatusNotification);
            }
            sleep(5);
        } else {
			// We are done.
            XgMessage msg;
			
            task->mStatus = kTaskTerminated;
            task->mTaskMegahertz = 0;
            
			if (task->mDataAvailable)
				encode_base64(task->mStandardOutput, task->mStandardOutputLength, &task->mBase64StandardOutput, &task->mBase64StandardOutputLength);
			
            TaskStatusNotificationMessage(&msg, task);
			MUTEX_UNLOCK(&task->mLock);

            if ( (err = xgridagent_SengMSG (simple, &error, msg.mStringRepresentation, strlen(msg.mStringRepresentation)) ) < 0) {
        		ReportError(kCouldNotSendStatusNotification);
        	}
            break;
        }
    }
	
}

void*
XgridTaskThread(void* inPtr)
{
    Task* task = (Task*)inPtr;
    char buf[MAX_MEDIUM_BUFFER];
    char args[MAX_MEDIUM_BUFFER];
	struct stat stdoutStatsAfter, tarredWorkDir;
    long i;
    long res;
	int err;
	char* stdoutFilenamePtr;
	char * wdirPtr;
	
	sprintf(buf,"Entering Xgrid task thread for task #%ld",task->mTaskRef);
	LogMessage(buf, kVerbose);
	
	/* Move to a temporary directory */
	if (chdir("/tmp/")) {
		ReportError(kCouldNotCreateTmpDirectories);
		return NULL;
	}
	wdirPtr = tmpnam(NULL);

	if (mkdir(wdirPtr,(0777))) {
		ReportError(kCouldNotCreateTmpDirectories);
		return NULL;
	}

	if (chdir(wdirPtr)) {
		ReportError(kCouldNotCreateTmpDirectories);
		return NULL;
	}
	LogMessage("Preparing working directory", kDebug);
	
	/* First lets untar working directory */
	
	char* decoded;
	long length;
	if (task->mBase64TarredWorkingDirectory != NULL) {
		decode_base64(task->mBase64TarredWorkingDirectory, strlen(task->mBase64TarredWorkingDirectory), &decoded, &length);
		
		FILE* exe;
		if ( (exe = fopen("TarredWorkingDirectory.tar.gz","w")) == NULL) {
		  LogMessage("Can't open Tarred working directory", kNormal);
		  sprintf(buf,"System error: %d",errno);
		  LogMessage(buf,kDebug);
			return NULL;
		}

		for (i = 0; i < length; i++) {
			fputc(decoded[i], exe);
		}
		fclose(exe);
		free(decoded);
		
		if  ( (err = system("tar xzvf TarredWorkingDirectory.tar.gz")) != 0) {
			LogMessage("Error untarring TarredWorkingDirectory.tar.gz",kDebug);
		}
	}

        LogMessage("Preparing executable", kDebug);
	/* and untar executable */
	if (task->mBase64TarredExecutable != NULL) {
		decode_base64(task->mBase64TarredExecutable, strlen(task->mBase64TarredExecutable), &decoded, &length);
		
		FILE* exe;
		if ( (exe = fopen("TarredExecutable.tar.gz","w")) == NULL) {
                  LogMessage("Can't open Tarred Executable directory", kNormal);
                  sprintf(buf,"System error: %d",errno);
                  LogMessage(buf,kDebug);

			return NULL;
		}
		for (i = 0; i < length; i++) {
			fputc(decoded[i], exe);
		}
		fclose(exe);
		free(decoded);
		
		sprintf(buf,"tar xzvf TarredExecutable.tar.gz -C %s/",task->mTarredWorkingDirectoryName);
		if  ( (err = system(buf)) != 0) {
			LogMessage("Error untarring TarredExecutable.tar.gz",kDebug);
		}
	}

	/* and move into it */
        LogMessage("Moving into working directory", kDebug);
	if (task->mBase64TarredWorkingDirectory != NULL) {
		if (chdir(task->mTarredWorkingDirectoryName)) {
			ReportError(kCouldNotCreateTmpDirectories);
			return NULL;
		}
	}

        LogMessage("Concatenate arguments", kDebug);
	/* Second put arguments together */
	for (i = 0; i < task->mNumberOfArguments; i++) {
		strcat(args," ");
		strcat(args, task->mArguments[i]);
	}

	/* Third, get the name of the file we will use for stdout and working directory */
	stdoutFilenamePtr = tmpnam(NULL);
	
	if (task->mNumberOfArguments == 0) {
		if (task->mStandardInput == NULL) {
			sprintf(buf, "%s > %s", task->mPathCommand, stdoutFilenamePtr);
		} else {
			sprintf(buf, "echo \"%s\" | %s > %s", task->mStandardInput, task->mPathCommand, stdoutFilenamePtr);
		}
	} else {
		if (task->mStandardInput == NULL) {
			sprintf(buf, "%s %s > %s",task->mPathCommand, args, stdoutFilenamePtr);
		} else {
			sprintf(buf, "echo \"%s\" | %s %s > %s",task->mStandardInput, args, task->mPathCommand, stdoutFilenamePtr);
		}
	}


	
	LogMessage("Executing actual task via system call", kVerbose);
	LogMessage(buf, kVerbose);    
    
	/* Here is the actual system call */

	res = system(buf);

	/* We really want all buffers to be written to disk (including the redirected stdout)  */
	fflush(NULL);

	MUTEX_LOCK(&task->mLock);
	task->mTerminationResult = res;
	
	/* Did we produce anything on standard output ? */
	err = stat(stdoutFilenamePtr, &stdoutStatsAfter);
	
	if (err == 0 ) {
		FILE* out;
		if ( (out = fopen(stdoutFilenamePtr,"r")) == NULL) {
			ReportError(kCouldNotOpenStdoutFile);
			MUTEX_UNLOCK(&task->mLock);
			return NULL;
		}

		if ( (task->mStandardOutput = malloc(stdoutStatsAfter.st_size + 1)) == NULL) {
			ReportError(kOutOfMemoryErr);
			MUTEX_UNLOCK(&task->mLock);
			return NULL;
		}
		
		task->mStandardOutputLength = stdoutStatsAfter.st_size;

		sprintf(gMessageString,"Size of stdout: %ld bytes",task->mStandardOutputLength);
		LogMessage(gMessageString, kVerbose);
		
		for (i = 0; i < task->mStandardOutputLength; i++) {
			task->mStandardOutput[i] = fgetc(out);
		}
		task->mStandardOutput[i] = 0;
		task->mDataAvailable = 1;

		fclose(out);
	} else {
		task->mDataAvailable = 0;
		task->mStandardOutputLength = 0;
		task->mStandardOutput = NULL;
		task->mBase64StandardOutputLength = 0;
		task->mBase64StandardOutput = NULL;
	}

	
	/* Finally, tar the working directory to ship it back to controller */
	if (chdir("../")) {
	  MUTEX_UNLOCK(&task->mLock);
	  return NULL;
	} else if ( strlen(task->mTarredWorkingDirectoryName) != 0) {
		sprintf(buf,"tar cvzf TarredDirectoryData.tar.gz %s",task->mTarredWorkingDirectoryName);
		LogMessage(buf,kDebug);

		if  ( (err = system(buf)) != 0) {
			LogMessage("Error tarring TarredDirectoryData.tar.gz",kDebug);
			LogMessage(task->mTarredWorkingDirectoryName,kDebug);
		}
		
		err = stat("TarredDirectoryData.tar.gz", &tarredWorkDir);
		if (err != 0 && err != -1 ) {
			ReportError(kCouldNotObtainStatsOnStdoutFile);
		}
		
		char * binaryTarredWorkingDirectory;
		long length = tarredWorkDir.st_size;
		if ( (binaryTarredWorkingDirectory = malloc(length+1))  == NULL) {
			ReportError( kOutOfMemoryErr );
			MUTEX_UNLOCK(&task->mLock);		
			return NULL;
		}
		
		FILE* workDir;
		if ( (workDir = fopen("TarredDirectoryData.tar.gz","r")) == NULL) {
			MUTEX_UNLOCK(&task->mLock);
			return NULL;
		}

		int c;
		long i =0;
		while ( (c = fgetc(workDir)) != EOF) {
			binaryTarredWorkingDirectory[i] = c;
			++i;
		}

		binaryTarredWorkingDirectory[i] = 0;
		length = i;
		fclose(workDir);
		
		free(task->mBase64TarredWorkingDirectory);
		task->mBase64TarredWorkingDirectoryLength = 0;
		
		encode_base64(binaryTarredWorkingDirectory, length, &(task->mBase64TarredWorkingDirectory), &(task->mBase64TarredWorkingDirectoryLength));
		free(binaryTarredWorkingDirectory);
		
	}

	MUTEX_UNLOCK(&task->mLock);
	
	SEM_POST(task->mTaskDone);
	LogMessage("Exiting Task thread", kDebug);
}


/* Process messsage, respond to it */

int
xgridagent_ProcessBeepMSG (char *ibuf, int len) {
	int err=kNoErr;
	LogMessage("Entering process message",kDebug);
	
	/* Fix for XML received: <?xml > must start on first line */
	long i=0;
	while (ibuf[i] != '<') {
		i++;
	}
	
	
	XgMessage msg;
	if ((msg.mStringRepresentation = malloc( (len-i+1) * sizeof(char))) != NULL) {
		memcpy (msg.mStringRepresentation, ibuf+i, len-i);
		msg.mStringRepresentation[len-i]=0;

		LogMessage("Message received", kVerbose);
		sprintf(gMessageString,"Current number of tasks: %d", gAgent.mCurrentTaskCount);
		LogMessage(gMessageString, kDebug);
		LogMessage(msg.mStringRepresentation, kDebugBEEPWithMessages);

	
		if ( (err = RecontructXgridXMLMessageFromText(&msg)) == kNoErr) {
			if (strcmp(msg.mName,"agentRegistration") == 0) {
				err = ProcessReplyToAgentRegistration(&msg);
			} else if (strcmp(msg.mName,"agentStatus") == 0) {
				err = ProcessReplyToAgentStatus(&msg);
			} else if (strcmp(msg.mName,"TaskSubmission") == 0) {
				err = ProcessTaskSubmission(&msg);
			} else if (strcmp(msg.mName,"TaskControl") == 0) {
				err = ProcessTaskControl(&msg);
			} else if (strcmp(msg.mName,"TaskResults") == 0) {
				err = ProcessTaskResults(&msg);
			} else if (strcmp(msg.mName,"TaskFinalResults") == 0) {
				err = ProcessTaskFinalResults(&msg);
			} else {
				err = ProcessUnknown(&msg);
			}
		}
	} else {
		err = kOutOfMemoryErr;
	}
	
	if(err != kNoErr) 
		ReportError(err);
	
	return kNoErr;
}

int ProcessReplyToAgentRegistration(XgMessage* inMessage)
{
	char buf[MAX_MEDIUM_BUFFER];
	int err = kNoErr;
	
	if ( !extractSingleXPath( inMessage->mXMLDocRepresentation, "/plist/dict/dict/key[1]/text()", buf) ) {
		if (strcmp("error",buf) == 0) {
		  return kCantGetCookie;
		} else if (strcmp("agentCookie",buf) == 0) {
			if ( !extractSingleXPath( inMessage->mXMLDocRepresentation, "/plist/dict/dict/string[preceding-sibling::key[position()=1 and text()='agentCookie']]/text()", buf) ) {
				LogMessage("Processing registration cookie", kVerbose);
				strcpy(gAgent.mAgentCookie,buf);
				FILE* cookie;
				if ( (cookie = fopen("cookie","w")) == NULL) {
				  LogMessage("Cannot save registration cookie.  You might need to restart the GridServer next time, or change the name of this agent.", kNormal);
				} else {
				  fprintf(cookie,"%s", gAgent.mAgentCookie);
				  fclose(cookie);
				}

			} else {
				err = kCantGetCookie;
			}
		}
	}
			 
	return err;
}


int ProcessReplyToAgentStatus(XgMessage* inMessage)
{
	char buf[MAX_MEDIUM_BUFFER];
	int err = kNoErr;
	GError *error = NULL;
	
	LogMessage("Processing reply status request", kVerbose);
	
	XgMessage msg;
	msg.mCorrelationID = inMessage->mCorrelationID;
	StatusReplyOrNotificationMessage(&msg, XgMessageReplyTypeValue);
	
    if ( (xgridagent_SengMSG (simple, &error, msg.mStringRepresentation, strlen(msg.mStringRepresentation)) ) < 0) {
		err = kCantSendBeepMessage;
	}
	
	return err;
}

int ProcessTaskSubmission(XgMessage* inMessage)
{
	char buf[MAX_MEDIUM_BUFFER];
	char base64stdin[MAX_MEDIUM_BUFFER];
	char *input;
	long length; 
	char command[MAX_MEDIUM_BUFFER];
	int err;
	GError *error = NULL;
	
	LogMessage("Processing task submission request", kVerbose);
	
	Task* task;
	task = NewXgridTask();
	
	while (gAgent.mCurrentTaskCount >= gAgent.mMaxTaskCount) {
		LogMessage("Waiting for tasks to finish and clean up", kVerbose);
		sleep(1);
	}

	XgMessage reply;
	reply.mCorrelationID = inMessage->mCorrelationID;
	
	if (gAgent.mCurrentTaskCount < gAgent.mMaxTaskCount) {
		if ( !extractSingleXPath( inMessage->mXMLDocRepresentation, "/plist/dict/dict/dict/data[preceding-sibling::key[position()=1 and text()='StandardInput']]/text()", base64stdin) ) {
			strcpyTrimmed(base64stdin, base64stdin);
			decode_base64(base64stdin, strlen(base64stdin), &task->mStandardInput, &task->mStandardInputLength);
		}
		
		if ( !extractSingleXPathInNewBuffer( inMessage->mXMLDocRepresentation, "/plist/dict/dict/dict/data[preceding-sibling::key[position()=1 and text()='TarredExecutable']]/text()", &(task->mBase64TarredExecutable), &(task->mBase64TarredExecutableLength) ) ) {
			strcpyTrimmed(task->mBase64TarredExecutable, task->mBase64TarredExecutable);
		}

		if ( !extractSingleXPathInNewBuffer( inMessage->mXMLDocRepresentation, "/plist/dict/dict/dict/data[preceding-sibling::key[position()=1 and text()='TarredWorkingDirectory']]/text()", &(task->mBase64TarredWorkingDirectory), &(task->mBase64TarredWorkingDirectoryLength) ) ) {
			strcpyTrimmed(task->mBase64TarredWorkingDirectory, task->mBase64TarredWorkingDirectory);
		}
		err = extractSingleXPath( inMessage->mXMLDocRepresentation, "/plist/dict/dict/dict/string[preceding-sibling::key[position()=1 and text()='TarredWorkingDirectoryName']]/text()", buf);
		if (err != kNoErr && err != kXPathResultIsEmpty) {
			return err;
		} else {
			strcpyTrimmed(task->mTarredWorkingDirectoryName, buf);
		}
		
		err = extractArrayOfStringsXPath( inMessage->mXMLDocRepresentation, "/plist/dict/dict/array[preceding-sibling::key[position()=1 and text()='XgTaskSpecificationArguments']]/string/text()", task->mArguments, &task->mNumberOfArguments); 
		if (err != kNoErr && err != kXPathResultIsEmpty) {
			return err;
		}
		
		if ( (err = extractSingleXPath( inMessage->mXMLDocRepresentation, "/plist/dict/dict/string[preceding-sibling::key[position()=1 and text()='XgTaskSpecificationPath']]/text()", buf)) != kNoErr ) {
			return err;
		} else {
			strcpyTrimmed(task->mPathCommand, buf);
		}

		if (AddXgridTaskToQueue(task)) {
			return kCantAddTaskToQueue;
		}
		
		TaskSubmissionReplyMessage(&reply, task);
		
	} else {
		strcpy(reply.mName,"TaskSubmission");
		reply.mType=XgMessageReplyTypeValue;
		reply.mXMLDocRepresentation = NULL;
		
		xmlNodePtr payloadDict;
		payloadDict = SetXgridXMLMessage(&reply);
		AddKeyStringPairAsChildren(payloadDict, BAD_CAST("error"),BAD_CAST("Unavailable"));

		LogMessage("Already running maximum number of tasks", kVerbose);
	}
	
    if ( (xgridagent_SengMSG (simple, &error, reply.mStringRepresentation, strlen(reply.mStringRepresentation)) ) < 0) {
		err = kCantSendBeepMessage;
	}
	
	return err;
}

int ProcessTaskControl(XgMessage* inMessage)
{
	char buf[MAX_MEDIUM_BUFFER];
	int cc, len;
	Task task;
	long taskref;
	GError *error = NULL;
	int err = kNoErr;
	
	LogMessage("Processing task control request", kVerbose);
	
	if ( !extractSingleXPath( inMessage->mXMLDocRepresentation, "/plist/dict/dict/string[preceding-sibling::key[position()=1 and text()='taskRef']]/text()", buf) ) {
		taskref = atol(buf);
	} else {
	    return kCouldNotExtractTaskRef;
	}
				
	if ( !extractSingleXPath( inMessage->mXMLDocRepresentation, "/plist/dict/dict/string[preceding-sibling::key[position()=1 and text()='control']]/text()", buf) ) {
		if (strcmp(buf,"Start") == 0) {
			
			if ( err = StartXgridTask(taskref) ) {
			    return err;
			}
			
			XgMessage reply;
			reply.mCorrelationID = taskref;
			
			TaskControlReplyMessage(&reply);
			
			if ( (cc = xgridagent_SengMSG (simple, &error, reply.mStringRepresentation, strlen(reply.mStringRepresentation)) ) < 0) {
				return kCantSendBeepMessage;
			}
		} else if (strcmp(buf,"cancel") == 0) {
			if ( !extractSingleXPath( inMessage->mXMLDocRepresentation, "/plist/dict/dict/string[preceding-sibling::key[position()=1 and text()='taskRef']]/text()", buf) ) {
				long taskRef = atol(buf);
				
				if ( CancelXgridTask(taskref) ) {
					return kCantCancelTask;
				}
				
				XgMessage reply;
				reply.mCorrelationID = inMessage->mCorrelationID;
				
				TaskControlReplyMessage(&reply);
				
				if ( (cc = xgridagent_SengMSG (simple, &error, reply.mStringRepresentation, strlen(reply.mStringRepresentation)) ) < 0) {
					return kCantSendBeepMessage;
				}
			}
		} else {
			return kUnknownTaskControlMessage;
		}
	}
	
	LogMessage("Exiting task control request", kVerbose);

	return kNoErr;
}

int ProcessTaskResults(XgMessage* inMessage)
{
	char buf[MAX_MEDIUM_BUFFER];
	int err;
	Task* task;
	long taskref;
	GError *error = NULL;
	
	sprintf(buf,"Processing task result request for task %ld",taskref);
	LogMessage(buf,kVerbose);

	
	if ( !extractSingleXPath( inMessage->mXMLDocRepresentation, "/plist/dict/dict/string[preceding-sibling::key[position()=1 and text()='taskRef']]/text()", buf) ) {
		taskref = atol(buf);
	} else {
	    return kCouldNotExtractTaskRef;
	}

	task = FindXgridTaskRef(taskref);

	MUTEX_LOCK(&task->mLock);
	
	XgMessage msg;
	msg.mCorrelationID = inMessage->mCorrelationID;
	
	TaskResultsReplyMessage(&msg, task);

	if ( (err = xgridagent_SengMSG (simple, &error, msg.mStringRepresentation, strlen(msg.mStringRepresentation)) ) < 0) {
		MUTEX_UNLOCK(&task->mLock);
		return kCantSendBeepMessage;
	}
	
	MUTEX_UNLOCK(&task->mLock);

	LogMessage("Exiting task result request", kVerbose);

	return  kNoErr;
	
}

int ProcessTaskFinalResults(XgMessage* inMessage)
{
	char buf[MAX_MEDIUM_BUFFER];
	int err, len;
	Task* task;
	long taskref;
	GError *error = NULL;
	
	LogMessage("Processing task final result request", kVerbose);
	
	if ( !extractSingleXPath( inMessage->mXMLDocRepresentation, "/plist/dict/dict/string[preceding-sibling::key[position()=1 and text()='taskRef']]/text()", buf) ) {
		taskref = atol(buf);
	} else {
		return kCouldNotExtractTaskRef;
	}
	
	task = FindXgridTaskRef(taskref);
	
	MUTEX_LOCK(&task->mLock);

	XgMessage msg;
	msg.mCorrelationID = inMessage->mCorrelationID;
	
	TaskFinalResultsReplyMessage(&msg, task);
	
	if ( (err = xgridagent_SengMSG (simple, &error, msg.mStringRepresentation, strlen(msg.mStringRepresentation)) ) < 0) {
		MUTEX_UNLOCK(&task->mLock);
		return kCantSendBeepMessage;
	}
	
	MUTEX_UNLOCK(&task->mLock);

	LogMessage("Exiting task final result request", kVerbose);

	return  kNoErr;
}

int ProcessUnknown(XgMessage* inMessage )
{
	LogMessage("Unknown XML message from controller.  Has Xgrid changed recently?", kNormal);
	
	LogMessage(inMessage->mStringRepresentation, kNormal);
	
	return kNoErr;
}



Task* NewXgridTask()
{
	Task* task = malloc( sizeof(Task) );
	long i;
	LogMessage("Creating Xgrid task",kDebug);
	
    if (task != NULL) {
      char buf[MAX_SMALL_BUFFER];
        task->mTaskRef = 0;
        task->mStandardInput = NULL;
        task->mStandardInputLength = 0;
		task->mBase64StandardInput = NULL;
		task->mBase64StandardInputLength = 0;
		
        task->mStandardOutput = NULL;
        task->mStandardOutputLength = 0;
		task->mBase64StandardOutput = NULL;
		task->mBase64StandardOutputLength = 0;
		
        task->mBase64TarredWorkingDirectory = NULL;
        task->mBase64TarredWorkingDirectoryLength = 0;
        task->mTarredWorkingDirectoryName[0] = 0;
		task->mBase64TarredExecutable = NULL;
		task->mBase64TarredExecutableLength = 0;
		
		for (i = 0; i < MAX_XGRID_ARGUMENTS; ++i) {
			task->mArguments[i] = NULL;
		}
	    task->mNumberOfArguments = 0;
		
        task->mPathCommand[0] = 0;
        task->mTaskMegahertz = gAgent.mMaximumCPUPower;
        task->mStatus = kTaskEmpty;
        task->mDataAvailable = 0;
        task->mTerminationResult = 0;
	task->mTaskRef = (++gAgent.mSerialTaskRef);
	sprintf(buf,"task%x",task);
	task->mTaskDone = sem_open(buf, O_CREAT,0777, 0);

        if (task->mTaskDone == SEM_FAILED) {
          sprintf(buf,"sem_open() failed with error number %d.  Trying sem_init().",errno);
          LogMessage(buf,kDebug);
          task->mTaskDone = malloc(sizeof(sem_t));
          if ( SEM_INIT(task->mTaskDone,0) ) {
            sprintf(buf,"Semaphore failed with error number %d",errno);
            LogMessage(buf,kDebug);
          }

	}


	

		if (task->mTaskDone == NULL) {
		  LogMessage("Can't create semaphore",kDebug);
		}
		pthread_mutex_init(&task->mLock, NULL );

		LogMessage("Exiting task creation",kDebug);
    } 
    
    return task;
}

void
DeleteXgridTask(Task* inTask)
{
	// Any pointer that is not NULL was allocated with malloc
	long i;

	LogMessage("Deleting Xgrid task",kDebug);

    MUTEX_LOCK(&inTask->mLock);
    free( inTask->mStandardInput );
    free( inTask->mBase64StandardInput );
    free( inTask->mStandardOutput );
    free( inTask->mBase64StandardOutput );
    free( inTask->mBase64TarredWorkingDirectory );  
    free( inTask->mBase64TarredExecutable );  
	
	for (i = 0; i < MAX_XGRID_ARGUMENTS; ++i) {
		free(inTask->mArguments[i]);
	}
	
	sem_close(inTask->mTaskDone);
    MUTEX_UNLOCK(&inTask->mLock);
	free(inTask);
}

int
AddXgridTaskToQueue(Task* inTask)
{
	long err = kNoErr;
	sprintf(gMessageString,"Adding Xgrid task #%d to queue (currently %d tasks)", inTask->mTaskRef, gAgent.mCurrentTaskCount);
	LogMessage(gMessageString,kDebug);

	if (gAgent.mCurrentTaskCount <= gAgent.mMaxTaskCount-1) {
		(gAgent.mCurrentTaskCount)++;
		gAgent.mTaskQueue[gAgent.mCurrentTaskCount-1] = inTask;
		sprintf(gMessageString,"Actually added task to queue (currently %d tasks)", gAgent.mCurrentTaskCount);
		LogMessage(gMessageString,kDebug);
	} else {
	  err = kTaskQueueIsFull;
	}

	return err;
}

int
RemoveXgridTaskFromQueue(long inTaskRef)
{
	long i,j;
	Task* task = NULL;
	sprintf(gMessageString,"Removing Xgrid task #%d from queue (currently %d tasks)", inTaskRef, gAgent.mCurrentTaskCount);
	LogMessage(gMessageString,kDebug);
	
	for (i = 0; i < gAgent.mCurrentTaskCount; i++) {
		task = gAgent.mTaskQueue[i];
		if (task->mTaskRef == inTaskRef) {
		  for (j = i; j < gAgent.mCurrentTaskCount-1; j++) {
		    gAgent.mTaskQueue[j]=gAgent.mTaskQueue[j+1];
		  }
		  (gAgent.mCurrentTaskCount)--;
		  sprintf(gMessageString,"Actually removed task from queue (currently %d tasks)", gAgent.mCurrentTaskCount);
		  LogMessage(gMessageString,kDebug);
		  break;
		}
	}
	
	if (task == NULL)
		return KCouldNotFindTaskRef;
	
	return kNoErr;
}

int
StartXgridTask(long inTaskRef)
{
	long i;
	Task* task = NULL;
	sprintf(gMessageString,"Signaling main thread to start Xgrid task #%d", inTaskRef);
	LogMessage(gMessageString,kDebug);

	for (i = 0; i < gAgent.mCurrentTaskCount; i++) {
		task = gAgent.mTaskQueue[i];
		if (task->mTaskRef == inTaskRef) {
			task->mStatus = kTaskMustStart;
			if ( SEM_POST(gAgent.mPendingTasks) ) {
			    return kCouldNotPostSemaphore;
			} else {
			    LogMessage("Signal sent succesfully", kDebug);
			}
		}
	}
	
	if (task == NULL)
		return KCouldNotFindTaskRef;
	
	return kNoErr;
}

Task*
FindXgridTaskRef(long inTaskRef)
{
	long i;
	Task* task = NULL;
	
	for (i = 0; i < gAgent.mCurrentTaskCount; i++) {
		task = gAgent.mTaskQueue[i];
		
		if (task->mTaskRef == inTaskRef) {
            break;
        }
	}
	
	return task;
}

int
CancelXgridTask(long inTaskRef)
{
	long i;
	Task* task = NULL;
	sprintf(gMessageString,"Signaling main thread to cancel Xgrid task #%d", inTaskRef);
	LogMessage(gMessageString,kDebug);

	for (i = 0; i < gAgent.mCurrentTaskCount; i++) {
		task = gAgent.mTaskQueue[i];
		MUTEX_LOCK(&task->mLock);
		if (task->mTaskRef == inTaskRef) {
			task->mStatus = kTaskMustCancel;
			SEM_POST(gAgent.mPendingTasks);
		}
		MUTEX_UNLOCK(&task->mLock);
	}	
	
	if (task == NULL)
		return KCouldNotFindTaskRef;
	
	return kNoErr;
	
}

int
TaskSubmissionReplyMessage(XgMessage* outMessage, Task * inTask) {
	
	xmlNodePtr payloadDict;
	char buf[MAX_SMALL_BUFFER];
	
	strcpy(outMessage->mName,"TaskSubmission");
	outMessage->mType=XgMessageReplyTypeValue;
	outMessage->mXMLDocRepresentation = NULL;
	
	payloadDict = SetXgridXMLMessage(outMessage);
	
	sprintf(buf,"%ld",inTask->mTaskRef);
	AddKeyStringPairAsChildren(payloadDict, BAD_CAST("taskRef"),BAD_CAST(buf));
	
	outMessage->mStringRepresentation = WriteXMLToTextBuffer(outMessage);
	
	return kNoErr;
}

int
TaskControlReplyMessage(XgMessage* outMessage) {
	
	xmlNodePtr payloadDict;
	char buf[MAX_SMALL_BUFFER];
	
	strcpy(outMessage->mName,"TaskControl");
	outMessage->mType=XgMessageReplyTypeValue;
	outMessage->mXMLDocRepresentation = NULL;
	
	payloadDict = SetXgridXMLMessage(outMessage);
	
	outMessage->mStringRepresentation = WriteXMLToTextBuffer(outMessage);
	
	return kNoErr;
}

int
TaskStatusNotificationMessage(XgMessage* outMessage, Task * inTask) {
	
	xmlNodePtr payloadDict;
	char buf[MAX_SMALL_BUFFER];
	LogMessage("Entering Task status", kDebug);
	
	strcpy(outMessage->mName,"TaskStatus");
	outMessage->mType=XgMessageNotificationTypeValue;
	outMessage->mXMLDocRepresentation = NULL;
	outMessage->mCorrelationID = inTask->mTaskRef;
	
	payloadDict = SetXgridXMLMessage(outMessage);
	
	AddKeyStringPairAsChildren(payloadDict, BAD_CAST("DataAvailable"),BAD_CAST(inTask->mDataAvailable == 0? "NO":"YES"));
	sprintf(buf,"%d",inTask->mTaskMegahertz);
	AddKeyStringPairAsChildren(payloadDict, BAD_CAST("TaskMegahertz"),BAD_CAST(buf));
	AddKeyStringPairAsChildren(payloadDict, BAD_CAST("TaskState"),BAD_CAST(gTaskStatusStrings[inTask->mStatus]));
    if (inTask->mStatus != kTaskRunning) {
    	sprintf(buf,"%d",inTask->mTerminationResult);
    } else {
    	buf[0]=0;
    }
	AddKeyStringPairAsChildren(payloadDict, BAD_CAST("TerminationStatus"),BAD_CAST(buf));
	
	outMessage->mStringRepresentation = WriteXMLToTextBuffer(outMessage);
	
	LogMessage("Exiting Task status", kDebug);

	return kNoErr;
}

int
TaskResultsReplyMessage(XgMessage* outMessage, Task * inTask) {
	
	xmlNodePtr payloadDict, taskResultDict, tarredDict;
	char buf[MAX_SMALL_BUFFER];
	
	LogMessage("Entering Task result reply setup",kDebug);
	
	strcpy(outMessage->mName,"TaskResults");
	outMessage->mType=XgMessageReplyTypeValue;
	outMessage->mXMLDocRepresentation = NULL;
	
	payloadDict = SetXgridXMLMessage(outMessage);
	
	if (xmlNewTextChild(payloadDict, NULL, "key", "TaskResults") == NULL) {
		return kCantAddXMLChild;
	}
	taskResultDict = xmlAddChild (payloadDict,xmlNewNode(NULL,BAD_CAST("dict")) );

	if ( xmlNewTextChild(taskResultDict, NULL,BAD_CAST("key"),BAD_CAST("ErrData") ) == NULL) {
		return kCantAddXMLChild;
	}
	
	if ( xmlNewTextChild(taskResultDict, NULL,BAD_CAST("data"),BAD_CAST("") ) == NULL) {
		return kCantAddXMLChild;
	}

	if ( xmlNewTextChild(taskResultDict, NULL,BAD_CAST("key"),BAD_CAST("ReadData") ) == NULL) {
	  return kCantAddXMLChild;
	}
		
	if ( (inTask->mBase64StandardOutputLength) != 0) {
	  if ( xmlNewTextChild(taskResultDict, NULL,BAD_CAST("data"),BAD_CAST(inTask->mBase64StandardOutput) ) == NULL) {
	    return kCantAddXMLChild;
	  }
	} else {
	  if ( xmlNewTextChild(taskResultDict, NULL,BAD_CAST("data"),BAD_CAST("") ) == NULL) {
	    return kCantAddXMLChild;
	  }
	}
	
	AddKeyStringPairAsChildren(payloadDict, BAD_CAST("agentName"),BAD_CAST(gAgent.mAgentName));
	sprintf(buf,"%d",inTask->mTaskRef);
	AddKeyStringPairAsChildren(payloadDict, BAD_CAST("taskRef"),BAD_CAST(buf));
	
	outMessage->mStringRepresentation = WriteXMLToTextBuffer(outMessage);
	
	return kNoErr;
}

int
TaskFinalResultsReplyMessage(XgMessage* outMessage, Task * inTask) {
	
	xmlNodePtr payloadDict, taskResultDict,tarredDict;
	char buf[MAX_SMALL_BUFFER];
	
	strcpy(outMessage->mName,"TaskFinalResults");
	outMessage->mType=XgMessageReplyTypeValue;
	outMessage->mXMLDocRepresentation = NULL;
	
	payloadDict = SetXgridXMLMessage(outMessage);
		
	if (xmlNewTextChild(payloadDict, NULL, "key", "TaskResults") == NULL) {
		return kCantAddXMLChild;
	}
	taskResultDict = xmlAddChild (payloadDict,xmlNewNode(NULL,BAD_CAST("dict")) );
	
	if (inTask->mBase64TarredWorkingDirectoryLength != 0) {
	  /*char * withLineBreaks;
		long i;
		long offset = 1;

		withLineBreaks = malloc(inTask->mBase64TarredWorkingDirectoryLength+inTask->mBase64TarredWorkingDirectoryLength/52+1);
		
		for (i = 0 ; i <inTask->mBase64TarredWorkingDirectoryLength; i++) { 
			withLineBreaks[i] = inTask->mBase64TarredWorkingDirectory[i];
			
			if (i % 52 == 0 && i != 0) {
				withLineBreaks[i+offset] = '\n';
				offset++;
			}
		}
		withLineBreaks[i+offset] = 0;
	  */
		xmlNewTextChild (taskResultDict,NULL,BAD_CAST("key"),BAD_CAST("TarredWorkingDirectory") );
		tarredDict = xmlAddChild (taskResultDict,xmlNewNode(NULL,BAD_CAST("dict")) );
		xmlNewTextChild (tarredDict,NULL,BAD_CAST("key"),BAD_CAST("TarredDirectoryData") );

		if ( xmlNewTextChild(tarredDict, NULL,BAD_CAST("data"),BAD_CAST(inTask->mBase64TarredWorkingDirectory) ) == NULL) {
			return kCantAddXMLChild;
		}
		AddKeyStringPairAsChildren(tarredDict,BAD_CAST("TarredDirectoryName"),BAD_CAST(inTask->mTarredWorkingDirectoryName));
		//		free(withLineBreaks); 

	}

	/*	if ( xmlNewTextChild(taskResultDict, NULL,BAD_CAST("key"),BAD_CAST("WorkingDirectoryFiles") ) == NULL) {
		return kCantAddXMLChild;
	}
	if ( xmlAddChild(taskResultDict,xmlNewNode(NULL,BAD_CAST("array"))) == NULL) {
		return kCantAddXMLChild;
	}
	*/

	AddKeyStringPairAsChildren(payloadDict, BAD_CAST("agentName"),BAD_CAST(gAgent.mAgentName));
	sprintf(buf,"%d",inTask->mTaskRef);
	AddKeyStringPairAsChildren(payloadDict, BAD_CAST("taskRef"),BAD_CAST(buf));
	
	outMessage->mStringRepresentation = WriteXMLToTextBuffer(outMessage);
	
	return kNoErr;
}


int ExtractXGridMessageParameters(xmlDocPtr doc, long* outCorrID, char* outName, long* outType)
{
	char buf[MAX_SMALL_BUFFER];
	// Extract Correlation ID
	if ( ! extractSingleXPath(doc, "/plist/dict/string[preceding-sibling::key[position()=1 and text()='correlationID']]/text()", buf) ) {
		*outCorrID = atol(buf);
	} else {
		return kCouldNotExtractCorrelationID;
	}
	
	// Extract name of message
	if ( ! extractSingleXPath(doc, "/plist/dict/string[preceding-sibling::key[position()=1 and text()='name']]/text()", buf) ) {
		strcpy(outName,buf);
	} else {
		strcpy(outName,"error");
		return kCouldNotExtractName;
	}
	
	// Extract type of message
    if ( ! extractSingleXPath(doc, "/plist/dict/string[preceding-sibling::key[position()=1 and text()='type']]/text()", buf) ) {
		if (strcmp("XgMessageReplyTypeValue",buf) == 0) {
			*outType = XgMessageReplyTypeValue;
		} else if ( strcmp("XgMessageRequestTypeValue", buf) == 0 ) {
			*outType = XgMessageRequestTypeValue;
		} else {
			*outType=XgMesssageUnknown;
			return kCouldNotExtractMessageType;
		}
	} else {
		*outType=XgMesssageUnknown;
		return kCouldNotExtractMessageType;
	}
	
	return kNoErr;
}

/* Return the message in response to status request.  msg.mCorrelationID must be set appropriately. */

int
StatusReplyOrNotificationMessage(XgMessage* outMessage, long inType) {
	xmlNodePtr payloadDict;
	xmlNodePtr taskArray;
	char buf[MAX_SMALL_BUFFER];
	long i;
	
	strcpy(outMessage->mName,"agentStatus");
	outMessage->mType=inType;
	outMessage->mXMLDocRepresentation = NULL;
	
	payloadDict = SetXgridXMLMessage(outMessage);
	
	sprintf(buf, "%ld", gAgent.mCurrentTaskCount);
	AddKeyStringPairAsChildren(payloadDict, BAD_CAST("CurrentTaskCount"),BAD_CAST(buf));
	xmlNewTextChild(payloadDict, NULL, "key", "TaskRefArray");
	taskArray = xmlAddChild (payloadDict,xmlNewNode(NULL,BAD_CAST("array")) );
	
	
	for (i = 0; i < gAgent.mCurrentTaskCount; i++) {
		sprintf(buf, "%ld", gAgent.mTaskQueue[i]->mTaskRef);
		if (xmlNewTextChild(taskArray, NULL, "string", buf) == NULL) {
			return kCantAddXMLChild;
		}
	}
	
	AddKeyStringPairAsChildren(payloadDict, BAD_CAST("agentCookie"),BAD_CAST(gAgent.mAgentCookie));
	AddKeyStringPairAsChildren(payloadDict, BAD_CAST("agentName"),BAD_CAST(gAgent.mAgentName));
	AddKeyStringPairAsChildren(payloadDict, BAD_CAST("agentStatus"),BAD_CAST(gAgentStatusStrings[gAgent.mAgentStatus]));
	sprintf(buf, "%ld", gAgent.mCurrentCPUPower);
	AddKeyStringPairAsChildren(payloadDict, BAD_CAST( "currentCPUPower"),BAD_CAST(buf));
	
	outMessage->mStringRepresentation = WriteXMLToTextBuffer(outMessage);
	
	return kNoErr;
	
}


int
RegistrationRequestMessage(XgMessage * outMessage) {
	xmlNodePtr payloadDict;
	char buf[MAX_SMALL_BUFFER];
	
	outMessage->mCorrelationID = 0;
	strcpy(outMessage->mName,"agentRegistration");
	outMessage->mType=XgMessageRequestTypeValue;
	outMessage->mXMLDocRepresentation = NULL;
	
	payloadDict = SetXgridXMLMessage(outMessage);
	
	AddKeyStringPairAsChildren(payloadDict, BAD_CAST("Address"),BAD_CAST(gAgent.mAddress));
	AddKeyStringPairAsChildren(payloadDict, BAD_CAST("Hostname"),BAD_CAST(gAgent.mHostname));
	AddKeyStringPairAsChildren(payloadDict, BAD_CAST("agentCookie"),BAD_CAST(gAgent.mAgentCookie));
	AddKeyStringPairAsChildren(payloadDict, BAD_CAST("agentName"),BAD_CAST(gAgent.mAgentName));
	sprintf(buf, "%ld",gAgent.mMaximumCPUPower);
	AddKeyStringPairAsChildren(payloadDict, BAD_CAST("maximumCPUPower"),BAD_CAST(buf));
	sprintf(buf, "%ld",gAgent.mMaxTaskCount);
	AddKeyStringPairAsChildren(payloadDict, BAD_CAST("maximumTaskCount"),BAD_CAST(buf));
	
	outMessage->mStringRepresentation = WriteXMLToTextBuffer(outMessage);
	
	return kNoErr;
}


/* Make XGrid XML document from message, returns xmlDocPtr in Message structure .
In addition, return Payload XML node ptr */

xmlNodePtr
SetXgridXMLMessage(XgMessage* inMessage)
{
	xmlNodePtr cur;
	xmlNodePtr thePlistNode,theDictNode,thePayloadDictNode;
	char buf[MAX_SMALL_BUFFER];
	
	inMessage->mXMLDocRepresentation = xmlNewDoc(BAD_CAST("1.0"));
	xmlCreateIntSubset(inMessage->mXMLDocRepresentation,BAD_CAST("plist"),BAD_CAST("-//Apple Computer//DTD PLIST 1.0//EN"),BAD_CAST("http://www.apple.com/DTDs/PropertyList-1.0.dtd"));
	
	thePlistNode = xmlNewNode(NULL,BAD_CAST("plist"));
	xmlDocSetRootElement(inMessage->mXMLDocRepresentation, thePlistNode);
	
	theDictNode = xmlNewNode(NULL,BAD_CAST("dict"));
	xmlAddChild (thePlistNode,theDictNode);
	
	sprintf(buf,"%ld",inMessage->mCorrelationID);
	AddKeyStringPairAsChildren(theDictNode, BAD_CAST("correlationID"),BAD_CAST(buf));
	AddKeyStringPairAsChildren(theDictNode, BAD_CAST("name"),BAD_CAST(inMessage->mName));
	
	xmlNewTextChild(theDictNode, NULL, BAD_CAST("key"), BAD_CAST("payload"));
	thePayloadDictNode = xmlNewNode(NULL,BAD_CAST("dict"));
	xmlAddChild (theDictNode,thePayloadDictNode);
	
	AddKeyStringPairAsChildren(theDictNode, BAD_CAST("type"),BAD_CAST(gMessageType[inMessage->mType]));
	
	return thePayloadDictNode;
}


int
AddKeyStringPairAsChildren(xmlNodePtr theParent, xmlChar * inKey, xmlChar * inString)
{
	if (xmlNewTextChild(theParent, NULL, "key", inKey) == NULL) {
		return kCantAddXMLChild;
	}
	
	if (xmlNewTextChild(theParent, NULL, "string", inString) == NULL) {
		return kCantAddXMLChild;
	}
	
	return kNoErr;
}


int
extractSingleXPath(xmlDocPtr doc, xmlChar* xpath, char* outString) {
	
	strcpy(outString,"");
	
	xmlNodeSetPtr nodeset;
	xmlXPathObjectPtr result;
	xmlChar *keyword;
	
	xmlXPathContextPtr context;
	
	if ( (context = xmlXPathNewContext(doc)) == NULL ) {
		return kCantAllocateXMLContext;
	}
	
	if ( (result = xmlXPathEvalExpression(xpath, context)) == NULL ) {
		return kCantProcessXpath;
	}
	
	if(xmlXPathNodeSetIsEmpty(result->nodesetval)){
		xmlXPathFreeObject (result);
		return kXPathResultIsEmpty;
	}
	xmlXPathFreeContext(context);
	
	if (result) {
		nodeset = result->nodesetval;
		
		if (nodeset->nodeNr == 1) {
			xmlNodePtr theNode = nodeset->nodeTab[0];
			
			if (theNode->children != NULL )
				strcpy(outString,theNode->children->content);
			else
				strcpy(outString,theNode->content);
		} else {
			return kTooManyNodesReturned;
		}
		xmlXPathFreeObject (result);
        xmlCleanupParser();
		
	}
	
	return kNoErr;
}

int
extractSingleXPathInNewBuffer(xmlDocPtr doc, xmlChar* xpath, char* *outBuffer, long* outLen) {
	
	xmlNodeSetPtr nodeset;
	xmlXPathObjectPtr result;
	xmlChar *keyword;
	
	xmlXPathContextPtr context;
	
	if ( (context = xmlXPathNewContext(doc)) == NULL ) {
		return kCantAllocateXMLContext;
	}
	
	if ( (result = xmlXPathEvalExpression(xpath, context)) == NULL ) {
		return kCantProcessXpath;
	}
	
	xmlXPathFreeContext(context);

	if(xmlXPathNodeSetIsEmpty(result->nodesetval)){
		xmlXPathFreeObject (result);
		return kXPathResultIsEmpty;
	}
	
	if (result) {
		nodeset = result->nodesetval;
		
		if (nodeset->nodeNr == 1) {
			xmlNodePtr theNode = nodeset->nodeTab[0];
			
			if (theNode->children != NULL ) {
				*outBuffer = malloc(strlen(theNode->children->content)*sizeof(char)+1);
				strcpy(*outBuffer,theNode->children->content);
			} else {
				*outBuffer = malloc(strlen(theNode->content)*sizeof(char)+1);
				strcpy(*outBuffer,theNode->content);
			}
		} else {
			return kTooManyNodesReturned;
		}
		xmlXPathFreeObject (result);
        xmlCleanupParser();
		
	}
	
	
	return kNoErr;
}


int
extractArrayOfStringsXPath(xmlDocPtr doc, xmlChar* xpath, char* * outArray, long * outHowMany) {
	
	xmlNodeSetPtr nodeset;
	xmlXPathObjectPtr result;
	xmlChar *keyword;
	long i;
	
	xmlXPathContextPtr context;
	
	if ( (context = xmlXPathNewContext(doc)) == NULL ) {
		return kCantAllocateXMLContext;
	}
	
	if ( (result = xmlXPathEvalExpression(xpath, context)) == NULL ) {
		return kCantProcessXpath;
	}

	xmlXPathFreeContext(context);
	
	if(xmlXPathNodeSetIsEmpty(result->nodesetval)){
		outArray = NULL;
		outHowMany = 0;
		xmlXPathFreeObject (result);
		return kXPathResultIsEmpty;
	}
	
	if (result) {
		nodeset = result->nodesetval;
		*outHowMany = nodeset->nodeNr;
		for ( i = 0; i < *outHowMany; ++i ) {
			xmlNodePtr theNode = nodeset->nodeTab[0];
			
			if (theNode->children != NULL ) {
				outArray[i] = malloc(strlen(theNode->children->content)*sizeof(char)+1);
				strcpy((outArray[i]),theNode->children->content);
			} else {
				outArray[i] = malloc(strlen(theNode->content)*sizeof(char)+1);
				strcpy((outArray[i]),theNode->content);
			}
		}
		
		xmlXPathFreeObject (result);
        xmlCleanupParser();
		
	} 
	
	return kNoErr;
}


/* Make XGrid XML document from text, returns xmlDocPtr in Message structure .
In addition, return Payload XML node ptr */

int
RecontructXgridXMLMessageFromText(XgMessage* inMessage)
{
	inMessage->mXMLDocRepresentation = (xmlDocPtr)xmlReadMemory(inMessage->mStringRepresentation, strlen(inMessage->mStringRepresentation), NULL, NULL, 0);
	
	if (inMessage->mXMLDocRepresentation == NULL) {
		ReportError (kCantParseXMLErr);
		LogMessage(inMessage->mStringRepresentation, kDebug);
		return kCantParseXMLErr;
	} else {
		long corrID, type;
		
		char buf[MAX_MEDIUM_BUFFER];
		// Extract Correlation ID
		if ( ! extractSingleXPath(inMessage->mXMLDocRepresentation, "/plist/dict/string[preceding-sibling::key[position()=1 and text()='correlationID']]/text()", buf) ) {
			inMessage->mCorrelationID = atol(buf);
		} else {
			LogMessage("Could not extract Correlation ID XPath", kDebug);
			return kCouldNotExtractCorrelationID;
		}
		
		// Extract name of message
		if ( ! extractSingleXPath(inMessage->mXMLDocRepresentation, "/plist/dict/string[preceding-sibling::key[position()=1 and text()='name']]/text()", buf) ) {
			strcpy(inMessage->mName,buf);
		} else {
			strcpy(inMessage->mName,"error");
			return kCouldNotExtractName;
		}
		
		// Extract type of message
		if ( ! extractSingleXPath(inMessage->mXMLDocRepresentation, "/plist/dict/string[preceding-sibling::key[position()=1 and text()='type']]/text()", buf) ) {
			if (strcmp(gMessageType[XgMessageRequestTypeValue],buf) == 0) {
				inMessage->mType = XgMessageRequestTypeValue;
			} else if ( strcmp(gMessageType[XgMessageReplyTypeValue], buf) == 0 ) {
				inMessage->mType = XgMessageReplyTypeValue;
			} else {
				inMessage->mType = XgMesssageUnknown;
				return kCouldNotExtractMessageType;
			}
		} else {
			inMessage->mType=XgMesssageUnknown;
			return kCouldNotExtractMessageType;
		}
	}
	
	return kNoErr;
}


/* Small utility functions */

/* Convert line breaks of a buffer.  Returned buffer (UTF-8) is zero terminated.
User responsible for freeing memory */

xmlChar*
ConvertLineBreaksToNewBuffer(xmlChar* inBuf, long inLength)
{
	xmlChar * tmp, *txt;
	tmp = (xmlChar*)xmlStrncatNew(BAD_CAST("\n"),BAD_CAST(inBuf),inLength);
	inLength++;
	
	if ( (txt=malloc(2 * (inLength+1) * sizeof(xmlChar))) == NULL) {
		return NULL;
	}
	
	long j = 0;
	long i = 0;
	for (i = 0; i <= inLength; i++) {
		if (tmp[i] == '\n') {
			txt[j]='\r';
			j++;
			txt[j]='\n';
		} else {
			txt[j]=tmp[i];	
		}
		j++;
	}
	
	return txt;
}

xmlChar*
WriteXMLToTextBuffer(XgMessage * inMessage)
{
	xmlChar *txt, *tmp;
	int len;
	xmlDocDumpFormatMemoryEnc(inMessage->mXMLDocRepresentation, &tmp, &len, NULL, 1);
	
	txt = ConvertLineBreaksToNewBuffer(tmp, len);
	
	free(tmp);
	
	return (xmlChar*)txt;
}

MyFrame * NewNode(MyFrame* inHeadNode, pthread_mutex_t* lock)
{
  
	MyFrame* n, *l;

	LogMessage("Entering newnode()",kDebug);
	if ( ( n = malloc(sizeof(MyFrame))) == NULL) {
		return n;
	}
	
	n->msg = -1;
	n->size=0;
	n->buf=NULL;
	n->prev=NULL;
	n->next=NULL;
	
	l = inHeadNode;

	if (l == NULL)
		return n;
	
	MUTEX_LOCK(lock);

	while (l->next != NULL) {
		l = l->next;
	}
	
	n->prev = l;
	l->next = n;
	MUTEX_UNLOCK(lock);
	
	LogMessage("Exiting newnode()",kDebug);
	return n;
	
}

char* AggregateAndRemoveMSG(MyFrame* inHeadNode, long inMessage, pthread_mutex_t* lock)
{
	MyFrame* n, *tmp;
	long size = 1;
	long offset = 0;
	char* buf;
	char messageString[MAX_MEDIUM_BUFFER];

	sprintf(messageString,"Aggregating BEEP MSG #%d into single buffer",inMessage);	
	LogMessage(messageString, kDebugBEEP);

	MUTEX_LOCK(lock);
	n = inHeadNode;
	while (n != NULL) {
		if (n->msg == inMessage) {
			size += n->size;
		}
		n = n->next;
	}
	
	if (size == 0) {
		LogMessage("Can't find BEEP message in linked list message", kDebugBEEP);
	} else {
		if ( (buf = malloc(size+1)) == NULL) {
			MUTEX_UNLOCK(lock);
			return NULL;
		}
	}
	
	n = inHeadNode;
	
	while (n != NULL) {
		if (n->msg == inMessage) {
			memcpy (buf+offset, n->buf, n-> size);
			offset+=n->size;

			if (n->prev != NULL) {
			  n->prev->next = n->next;
			}

			if (n->next != NULL) {
			  n->next->prev = n->prev;
			}
			tmp = n->next;
			free(n->buf);
			free(n);
			n = tmp;
		} else {
			n = n->next;
		}
	}
	MUTEX_UNLOCK(lock); 

	buf[offset] = 0;
	
	return buf;
}


/* Base64 encoding (transform three 8-bit numbers into four 6-bit numbers (from 0 to 63)
then map this number to a letter A-Z a-z,0-9 + and -.
When the number of characters to be encoded is not a multiple of three, pad with '='.
*/
char *encode = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
int decode[256] = { -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, 62, -1, -1, -1, 63, 52, 53,
	54, 55, 56, 57, 58, 59, 60, 61, -1, -1,
	-1, 512,-1, -1, -1, 0,  1,  2,  3,  4,
	5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
	15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 
	25, -1, -1, -1, -1, -1, -1, 26, 27, 28,
	29, 30, 31, 32, 33, 34, 35, 36, 37,38,
	39, 40, 41, 42, 43, 44, 45, 46, 47,48,
	49, 50, 51, -1, -1, -1, -1, -1, -1,-1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1
};

/* Encodes a buffer (buf) of length l into a newly created buffer outBuffer. Final length of buffer is returned
through outLength.  User is responsible for freeing outBuffer.
*/

int
encode_base64(unsigned char * buf, long l, char* * outBuffer, long *outLength)
{
	long i,j;
	
	short c1,c2,c3;
	long b1,b2,b3,b4;
	
	*outLength = l % 3 == 0 ? l / 3 * 4 : (l/3 + 1) * 4; 
	if ( (*outBuffer = (char *)malloc(*outLength + 1)) == NULL ) {
		return kOutOfMemoryErr;
	}
	
	j = 0;
	for (i = 0; i < l; i+=3) {
		c1 = buf[i];
		if ( i+1 < l)
			c2 = buf[i+1];
		else
			c2 = 512;
		
		if ( i+2 < l)
			c3 = buf[i+2];
		else
			c3 = 512;
		
		b1 = c1 >> 2;
		b2 = ((c1 & 0x03) << 4) | ((c2 & 0xf0) >> 4);
		b3 = ((c2 & 0x0f) << 2) | ((c3 & 0xc0) >> 6);
		b4 = (c3 & 0x3f);
		
		(*outBuffer)[j++] = encode[b1];
		(*outBuffer)[j++] = encode[b2];
		(*outBuffer)[j++] = c2 != 512 ? encode[b3]: '=';
		(*outBuffer)[j++] = c3 != 512 ? encode[b4]: '=';
	}
	(*outBuffer)[j] = 0;

	if (j != *outLength) {
	  LogMessage("j is not calculated length",kDebug);
	  *outLength=j;
	}

	return kNoErr;
}

/* Decodes a buffer (buf) of length l into a newly created buffer outBuffer. Final length of buffer is returned
through outLength.  User is responsible for freeing outBuffer.
*/

int
decode_base64(unsigned char * buf, long l, char* * outBuffer, long *outLength)
{
	long i,j;
	
	unsigned short c1,c2,c3;
	long b1,b2,b3,b4;
	
	if (l % 4 == 0) {
		*outLength = l / 4 * 3;
	} else {
		return kIncorrectBase64Length;
	}
	
	if ( (*outBuffer = (char *)malloc(*outLength)) == NULL ) {
		return kOutOfMemoryErr;
	}
	
	j = 0;
	for (i = 0; i < l; i+=4) {
		if ( (b1 = decode[buf[i]]) == -1)
			return kInvalidBase64Character;
		if ( (b2 = decode[buf[i+1]]) == -1)
			return kInvalidBase64Character;
		if ( (b3 = decode[buf[i+2]]) == -1)
			return kInvalidBase64Character;
		if ( (b4 = decode[buf[i+3]]) == -1)
			return kInvalidBase64Character;
		
		c1 = b1 << 2 | ((b2 & 0x30) >> 4);
		c2 = ((b2 & 0x0f) << 4) | ((b3 & 0x3c) >> 2);
		c3 = ((b3 & 0x03) << 6) | (b4 & 0x3f);
		
		if (b4 == 512) {
			if (i+3 == l-1) {
				(*outLength)--;
			} else {
				return kInvalidBase64Character; 
			}
		}
		if (b3 == 512) {
			if (i+3 == l-1 && b4 == 512) {
				(*outLength)--;
			} else {
				return kInvalidBase64Character; 
			}
		}
		
		(*outBuffer)[j++] = c1;
		(*outBuffer)[j++] = c2;
		(*outBuffer)[j++] = c3;
	}
	(*outBuffer)[*outLength]=0;
	
	return kNoErr;
}

// Dest can be src
void
strcpyTrimmed(char * dest, char* src)
{
	long i,j;
	
	j = 0;
	for (i = 0; i < strlen(src); i++) {
		if (src[i] == ' ' || src[i] == '\t' || src[i] == '\n' || src[i] == '\r')
			continue;
		dest[j]=src[i];
		++j;
	}
	dest[j]=0;
}
