
#ifndef __XGRIDAGENT_H
#define __XGRIDAGENT_H

#include <pthread.h>
#include <semaphore.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>

#define DEFAULT_CONTENT_TYPE    "Content-Type: application/beep+xml\n\n"

/* Some unfortunate hardcoded limits */

#define MAX_SMALL_BUFFER 100
#define MAX_MEDIUM_BUFFER 1000
#define MAX_HUGE_BUFFER 10000000
#define MAX_XGRID_ARGUMENTS 100
#define MAX_THREADS 1000

/* Agent status */
enum { kAgentAvailable = 1};

/* Task status and actions */
enum {  
    kTaskRunning = 1,
	kTaskStopped = 2,
	kTaskTerminated = 3,
	kTaskMustCancel = 4,
	kTaskMustStart = 5,
    kTaskEmpty = 6
};

enum {
	XgMesssageUnknown = 0,
	XgMessageRequestTypeValue = 1,
	XgMessageReplyTypeValue = 2,
	XgMessageNotificationTypeValue = 3
} XgMessageTypeValue;

enum { 
	kNoErr = 0,
	kOutOfMemoryErr,
	kCantParseXMLErr,
	kCouldNotStartChannel,
	kCouldNotSendRegistration,
	kCouldNotCloseChannel,
	kCouldNotStartMainThread,
	kCouldNotStartTaskThread,
	kCouldNotSendStatusNotification,
	kCouldNotObtainStatsOnStdoutFile,
	kCouldNotOpenStdoutFile,
	kCantGetCookie,
	kTaskQueueIsFull,
	KCouldNotFindTaskRef,
	kIncorrectBase64Length,
	kInvalidBase64Character,
	kCouldNotExtractCorrelationID,
	kCouldNotExtractName,
	kCouldNotExtractMessageType,
	kCouldNotExtractTaskRef,
	kCantAllocateXMLContext,
	kCantCancelTask,
	kCantProcessXpath,
	kXPathResultIsEmpty,
	kTooManyNodesReturned,
	kCantAddXMLChild,
	kCantAddTaskToQueue,
	kCantSendBeepMessage,
	kCantOpenFile,
	kUnknownTaskControlMessage,
	kMaximumNumberOfTaskReached,
	kCouldNotCreateTmpDirectories,
	kCantReadConfigFile,
	kMustHaveTaskCountPositive,
	kMustHaveCPUPowerPositive,
	kCouldNotReadMaxTaskCount,
	kCouldNotReadServiceName,
	kCouldNotReadCPUPower,
	kCouldNotReadMessageLevel,
	kCouldNotGetAddress,
	kMustHavePositiveMessageLevel,
	kCouldNotPostSemaphore
};

enum {
	kNormal,
	kVerbose,
	kDebug,
	kDebugBEEP,
	kDebugBEEPWithMessages
};

/* XGrid message structure */
typedef struct {
	long mCorrelationID;
	char mName[MAX_SMALL_BUFFER];
	long mType;
	xmlDocPtr mXMLDocRepresentation;
	char* mStringRepresentation;
} XgMessage;

/* Task structure */
typedef struct {
	long mTaskRef;
	
	char *mStandardInput;
	long mStandardInputLength;
	char *mBase64StandardInput;
	long mBase64StandardInputLength;
	
	char *mStandardOutput;
	long mStandardOutputLength;
	char *mBase64StandardOutput;
	long mBase64StandardOutputLength;
	
	char *mBase64TarredWorkingDirectory;
	long mBase64TarredWorkingDirectoryLength;
	char mTarredWorkingDirectoryName[MAX_SMALL_BUFFER];
	
	char *mBase64TarredExecutable;
	long mBase64TarredExecutableLength;
	
	char* mArguments[MAX_XGRID_ARGUMENTS];
	long mNumberOfArguments;
	char mPathCommand[MAX_MEDIUM_BUFFER];
	
	long mTaskMegahertz;
	long mStatus;
	long mDataAvailable;
	long mTerminationResult;
    sem_t* mTaskDone;
    pthread_mutex_t mLock;
  	
} Task;

/* Agent structure */
typedef struct {
	long mMaxTaskCount;
	long mCurrentTaskCount;
	char mAgentName[MAX_SMALL_BUFFER];
	char mAddress[MAX_SMALL_BUFFER];
	char mHostname[MAX_SMALL_BUFFER];
	long mAgentStatus;
	char mAgentCookie[MAX_SMALL_BUFFER];
	long mCurrentCPUPower;
	long mMaximumCPUPower;
	long mMessageLevel;
	
	Task** mTaskQueue;
	long mSerialTaskRef;
	
	sem_t* mPendingTasks;
	pthread_mutex_t mXgridQueueLocked;
	
} Agent;


int
xgridagent (RRSimple* simple, GError** error);

int ExtractXGridMessageParameters(xmlDocPtr doc, long* outCorrID, char* outName, long* outType);
xmlChar*	RegistrationRequestString();
int		StatusReplyOrNotificationMessage(XgMessage* outMessage, long inType);
int ExtractXGridMessageParameters(xmlDocPtr doc, long* outCorrID, char* outName, long* outType);
int	xgridagent_ProcessBeepMSG (char *ibuf, int len);
int ProcessReplyToAgentRegistration(XgMessage* inMessage);
int ProcessReplyToAgentStatus(XgMessage* inMessage);
int ProcessTaskSubmission(XgMessage* inMessage);
int ProcessUnknown(XgMessage* inMessage );
int TaskSubmissionReplyMessage(XgMessage* outMessage, Task * inTask);
int TaskControlReplyMessage(XgMessage* outMessage);

int extractSingleXPathInNewBuffer(xmlDocPtr doc, xmlChar* xpath, char* *outBuffer, long * outLen);
int extractSingleXPath(xmlDocPtr doc, xmlChar* xpath, char* outString);
int extractArrayOfStringsXPath(xmlDocPtr doc, xmlChar* xpath, char* * outArray, long * outHowMany);

int AddKeyStringPairAsChildren(xmlNodePtr theParent, xmlChar * inKey, xmlChar * inString);
void	 strcpyTrimmed(char * dest, char* src);
xmlChar* ConvertLineBreaksToNewBuffer(xmlChar* inBuf, long inLength);
xmlChar* WriteXMLToTextBuffer(XgMessage * inMessage);
xmlNodePtr SetXgridXMLMessage(XgMessage* inMessage);
int RecontructXgridXMLMessageFromText(XgMessage* inMessage);

int AddXgridTaskToQueue(Task* inTask);
int RemoveXgridTaskFromQueue(long inTaskRef);
int StartXgridTask(long inTaskRef);
int CancelXgridTask(long taskref);

Task* NewXgridTask();
void DeleteXgridTask(Task* inTask);
Task* FindXgridTaskRef(long inTaskRef);

void* SimpleWatchThread(void * inPtr);
void* XgridTaskMonitorThread(void* inTask);
void* XgridTaskThread(void* inTask);

int ReportError(int inErr);
void LogMessage(char* inMsg, long inLevel);

MyFrame * NewNode(MyFrame* inHeadNode, pthread_mutex_t* lock);
char* AggregateAndRemoveMSG(MyFrame* inHeadNode, long inMessage, pthread_mutex_t* lock);

RRSimple* GetSimple();

static char * gMessageType[] = {"none","XgMessageRequestTypeValue","XgMessageReplyTypeValue","XgMessageNotificationTypeValue"};
static char * gAgentStatusStrings[] = {"none","Available"};
static char * gTaskStatusStrings[] = {"none","Running","Stopped","Terminated","MustCancel","MustStart","Empty"};
static char * gErrorStrings[] = { 
	"none",
	"Out of memory",
	"Cannot parse XML",
	"Could not start BEEP channel",
	"Could not send registration",
	"Could not close BEEP channel",
	"Could not start main thread",
	"Could not start task thread",
	"Could not send status notification",
	"Could not obtain information on standard output file",
	"Could not open standard output file",
	"Cannot get agent registration cookie",
	"Task queue is full",
	"Could not find task in task queue",
	"Incorrect base 64 length encoding",
	"Invalid base 64 encoding character",
	"Could not extract Xgrid correlation ID",
	"Could not extract Xgrid message name",
	"Could not extract Xgrid message type",
	"Could not extract Xgrid Task reference",
	"Could not allocate XML context",
	"Cannot cancel task",
	"Cannot process Xpath",
	"Xpath result is empty",
	"Too many nodes returned",
	"Cannot add XML child",
	"Cannot add task to Xgrid task queue",
	"Cannot send BEEP message",
	"Cannot open file",
	"Unknown task control message",
	"Maximum number of task reached",
	"Cannot create temporary directory",
	"Cannot read configuration file",
	"The maximum number of task in configuration file must be more than 0",
	"The maximum CPU power in configuration file must be more than 0",
	"Cannot read maximum task count in configuration file",
	"Cannot read service name in configuration file",
	"Cannot read CPU power in configuration file",
	"Cannot read message level in configuration file",
	"Cannot get IP address of this agent",
	"Message level in configuration file must be more than 0",
	"Could not post semaphore to start task"
};

#define SEM_INIT(x, count) sem_init(x,0, count)
#define SEM_POST(x) sem_post(x)
#define SEM_WAIT(x) sem_wait(x)
#define SEM_TRYWAIT(x) sem_trywait(x)
#define MUTEX_LOCK(x) {char buf[100]; sprintf(buf,"Locking mutex %x",x); LogMessage(buf, kDebug); pthread_mutex_lock(x);}
#define MUTEX_UNLOCK(x) {char buf[100]; sprintf(buf,"Unlocking mutex %x",x); LogMessage(buf, kDebug); pthread_mutex_unlock(x);}

#endif
